# -*- coding: utf-8 -*-

from collective.timestamp import _
from collective.timestamp import logger
from collective.timestamp.behaviors.timestamp import ITimestampableDocument
from collective.timestamp.interfaces import ITimeStamper
from plone import api
from Products.Five.browser import BrowserView
from rfc3161ng import TimestampingError


class TimestampView(BrowserView):

    def available(self):
        """
        Show timestamp action only if content is stampable and not already
        stamped.
        """
        if not ITimestampableDocument.providedBy(self.context):
            return False
        handler = ITimeStamper(self.context)
        return handler.is_timestampable()

    def timestamp(self):
        obj = self.context
        redirect_url = f"{obj.absolute_url()}/view"
        handler = ITimeStamper(obj)
        try:
            handler.timestamp()
        except TimestampingError as e:
            api.portal.show_message(
                _("Timestamp has failed."),
                self.request,
                type="error",
            )
            logger.error(f"Timestamp action failed for {obj.absolute_url()} : {str(e)}")
            self.request.response.redirect(redirect_url)
            return ""
        logger.info(f"Timestamp generated for {obj.absolute_url()}")
        api.portal.show_message(
            _("Timestamp file has been successfully generated and saved"), self.request
        )
        self.request.response.redirect(redirect_url)



# -*- coding: utf-8 -*-

from collective.timestamp.interfaces import ITimeStamper
from collective.timestamp.testing import COLLECTIVE_TIMESTAMP_INTEGRATION_TESTING
from plone import api
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.namedfile.file import NamedBlobFile
from Products.statusmessages.interfaces import IStatusMessage
from rfc3161ng import TimestampingError
from unittest.mock import patch
from zope.lifecycleevent import modified

import unittest


class TestRule(unittest.TestCase):

    layer = COLLECTIVE_TIMESTAMP_INTEGRATION_TESTING

    def setUp(self):
        """Custom shared utility setup for tests."""
        self.request = self.layer["request"]
        self.portal = self.layer["portal"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.file = api.content.create(
            container=self.portal,
            type="File",
            id="my-file",
        )

    def test_rule_action(self):
        handler = ITimeStamper(self.file)
        self.assertFalse(handler.is_timestamped())
        self.file.file = NamedBlobFile(data=b"file data", filename="file.pdf")
        modified(self.file)
        self.assertTrue(handler.is_timestamped())
        messages = IStatusMessage(self.request)
        show = messages.show()
        self.assertEqual(len(show), 1)
        self.assertEqual(
            "Timestamp file has been successfully generated and saved for my-file",
            show[0].message,
        )

    def test_rule_action_error(self):
        with patch(
            "collective.timestamp.adapters.TimeStamper.timestamp",
            side_effect=TimestampingError,
        ):
            self.file.file = NamedBlobFile(data=b"file data", filename="file.pdf")
            modified(self.file)
            messages = IStatusMessage(self.request)
            show = messages.show()
            self.assertEqual(len(show), 1)
            self.assertIn("Unable to timestamp", show[0].message)



# -*- coding: utf-8 -*-

from collective.timestamp.browser.viewlet import TimestampViewlet
from collective.timestamp.interfaces import ITimeStamper
from collective.timestamp.testing import COLLECTIVE_TIMESTAMP_INTEGRATION_TESTING
from plone import api
from plone.app.testing import logout
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.namedfile.file import NamedBlobFile
from zope.component import getMultiAdapter
from zope.interface.interfaces import ComponentLookupError

import unittest


class TestVerification(unittest.TestCase):

    layer = COLLECTIVE_TIMESTAMP_INTEGRATION_TESTING

    def setUp(self):
        """Custom shared utility setup for tests."""
        self.request = self.layer["request"]
        self.portal = self.layer["portal"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.document = api.content.create(
            container=self.portal,
            type="Document",
            id="my-document",
        )
        self.file = api.content.create(
            container=self.portal,
            type="File",
            id="my-file",
        )

    def test_viewlet(self):
        logout()
        viewlet = TimestampViewlet(self.document, self.request, None)
        with self.assertRaises(TypeError):
            viewlet.available()
        viewlet = TimestampViewlet(self.file, self.request, None)
        self.assertFalse(viewlet.available())
        self.file.file = NamedBlobFile(data=b"file data", filename="file.txt")
        self.assertFalse(viewlet.available())
        handler = ITimeStamper(self.file)
        handler.timestamp()
        self.assertTrue(viewlet.available())
        viewlet.update()
        html = viewlet.render()
        self.assertIn("svg", html)

    def test_view(self):
        logout()
        with self.assertRaises(ComponentLookupError):
            view = getMultiAdapter((self.document, self.request), name="timestamp")
        view = getMultiAdapter((self.file, self.request), name="timestamp")
        self.assertIn("This content is not timestamped.", view())
        self.assertFalse(view.is_timestamped())
        self.file.file = NamedBlobFile(data=b"file data", filename="file.txt")
        self.assertIn("This content is not timestamped.", view())
        self.assertFalse(view.is_timestamped())
        handler = ITimeStamper(self.file)
        handler.timestamp()
        self.assertIn("This content was timestamped on", view())
        self.assertEqual(
            view.more_infos_url(), "http://documentation.timestamptest.com"
        )
        self.assertTrue(view.is_timestamped())
        self.assertEqual(view.timestamp_date(), self.file.effective().asdatetime())



# -*- coding: utf-8 -*-

from collective.timestamp import _
from plone import schema
from plone.autoform.directives import read_permission
from plone.autoform.directives import write_permission
from plone.autoform.interfaces import IFormFieldProvider
from plone.namedfile.field import NamedBlobFile
from plone.supermodel import model
from zope.interface import provider


@provider(IFormFieldProvider)
class ITimestampableDocument(model.Schema):
    """ """

    model.fieldset(
        "timestamp",
        label=_("Timestamp"),
        fields=["enable_timestamping", "timestamp"],
    )

    read_permission(timestamp="collective.timestamp.read")
    write_permission(timestamp="collective.timestamp.write")
    timestamp = NamedBlobFile(
        title=_("Time Stamp Response (TSR) file"),
        required=False,
    )

    enable_timestamping = schema.Bool(
        title=_("Enable timestamping"),
        description=_(
            "This allows you to disable timestamping for this specific content."
        ),
        default=True,
        required=False,
    )



# -*- coding: utf-8 -*-

from collective.timestamp.interfaces import ITimeStamper
from collective.timestamp.testing import COLLECTIVE_TIMESTAMP_INTEGRATION_TESTING
from datetime import datetime
from plone import api
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.namedfile.file import NamedBlobFile
from plone.uuid.interfaces import IUUID

import unittest


class TestAdapter(unittest.TestCase):

    layer = COLLECTIVE_TIMESTAMP_INTEGRATION_TESTING

    def setUp(self):
        """Custom shared utility setup for tests."""
        self.request = self.layer["request"]
        self.portal = self.layer["portal"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.document = api.content.create(
            container=self.portal,
            type="Document",
            id="my-document",
        )
        self.file = api.content.create(
            container=self.portal,
            type="File",
            id="my-file",
        )

    def test_adapter(self):
        with self.assertRaises(TypeError):
            ITimeStamper(self.document)
        self.assertIsNotNone(ITimeStamper(self.file))

    def test_get_file_field(self):
        handler = ITimeStamper(self.file)
        self.assertIsNone(handler.get_file_field())
        self.file.file = NamedBlobFile(data=b"file data", filename="file.txt")
        field = handler.get_file_field()
        self.assertIsNotNone(field)
        self.assertEqual(field.fieldname, "file")

    def test_get_data(self):
        handler = ITimeStamper(self.file)
        self.assertIsNone(handler.get_data())
        self.file.file = NamedBlobFile(data=b"file data", filename="file.txt")
        self.assertEqual(handler.get_data(), b"file data")

    def test_is_timestamped(self):
        self.file.file = NamedBlobFile(data=b"file data", filename="file.txt")
        handler = ITimeStamper(self.file)
        self.assertFalse(handler.is_timestamped())
        handler.timestamp()
        self.assertTrue(handler.is_timestamped())

    def test_is_timestampable(self):
        handler = ITimeStamper(self.file)
        self.assertFalse(handler.is_timestampable())
        self.file.file = NamedBlobFile(data=b"file data", filename="file.txt")
        self.assertTrue(handler.is_timestampable())
        self.file.enable_timestamping = False
        self.assertFalse(handler.is_timestampable())
        self.file.enable_timestamping = True
        self.assertTrue(handler.is_timestampable())
        handler.timestamp()
        self.assertFalse(handler.is_timestampable())

    def test_timestamp(self):
        uuid = IUUID(self.file)
        handler = ITimeStamper(self.file)
        first_effective_date = self.file.effective()
        with self.assertRaises(ValueError):
            handler.timestamp()
        self.assertIsNone(self.file.timestamp)
        self.file.file = NamedBlobFile(data=b"file data", filename="file.txt")
        catalog = api.portal.get_tool("portal_catalog")
        brain = api.content.find(UID=uuid)[0]
        indexes = catalog.getIndexDataForRID(brain.getRID())
        first_effective_index = indexes.get("effective")
        first_effective_range_index = indexes.get("effectiveRange")
        self.assertFalse(indexes.get("is_timestamped"))
        data, timestamp_date = handler.timestamp()
        self.assertIsInstance(data, bytes)
        self.assertIsInstance(timestamp_date, datetime)
        self.assertEqual(data, self.file.file.data)
        self.assertEqual(timestamp_date, self.file.effective().asdatetime())
        self.assertIsNotNone(self.file.timestamp)
        self.assertNotEqual(first_effective_date, self.file.effective())
        brain = api.content.find(UID=uuid)[0]
        indexes = catalog.getIndexDataForRID(brain.getRID())
        self.assertNotEqual(first_effective_index, indexes.get("effective"))
        self.assertNotEqual(first_effective_range_index, indexes.get("effectiveRange"))
        self.assertTrue(indexes.get("is_timestamped"))



# -*- coding: utf-8 -*-

from collective.timestamp.interfaces import ITimeStamper
from collective.timestamp.interfaces import ITimestampingSettings
from collective.timestamp.utils import get_timestamp_date_from_tsr_file
from plone import api
from Products.Five.browser import BrowserView


class VerificationView(BrowserView):

    def is_timestamped(self):
        handler = ITimeStamper(self.context)
        return handler.is_timestamped()

    def timestamp_date(self):
        tsr_file = self.context.timestamp
        timestamp_date = get_timestamp_date_from_tsr_file(tsr_file.data)
        return timestamp_date.astimezone()

    def more_infos_url(self):
        return api.portal.get_registry_record(
            "timestamping_documentation_url",
            interface=ITimestampingSettings,
        )



# -*- coding: utf-8 -*-

from collective.timestamp.testing import COLLECTIVE_TIMESTAMP_INTEGRATION_TESTING
from collective.timestamp.utils import timestamp
from collective.timestamp.utils import get_timestamp_date_from_tsr_file
from collective.timestamp.utils import localize_utc_date
from datetime import datetime, timedelta
from plone import api
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.namedfile.file import NamedBlobFile

import unittest
import pytz


class TestUtils(unittest.TestCase):

    layer = COLLECTIVE_TIMESTAMP_INTEGRATION_TESTING

    def setUp(self):
        """Custom shared utility setup for tests."""
        self.request = self.layer["request"]
        self.portal = self.layer["portal"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.my_file = api.content.create(
            container=self.portal,
            type="File",
            id="my-file",
        )
        self.my_file.file = NamedBlobFile(data=b"file data", filename="file.txt")
        self.file_data = self.my_file.file.data

    def test_localize_utc_date(self):
        naive_datetime = datetime(2024, 9, 10, 12, 0, 0)
        localized_datetime = localize_utc_date(naive_datetime)
        expected_datetime = datetime(2024, 9, 10, 12, 0, 0, tzinfo=pytz.UTC)
        self.assertEqual(localized_datetime, expected_datetime)

    def test_timestamp(self):
        tsr, timestamp_date = timestamp(self.file_data, "http://freetsa.org/tsr")
        self.assertIsInstance(tsr, bytes)
        self.assertIsInstance(timestamp_date, datetime)
        self.assertAlmostEqual(
            timestamp_date, datetime.now(pytz.UTC), delta=timedelta(seconds=10)
        )

    def test_timestamp_retries(self):
        tsr, timestamp_date = timestamp(
            self.file_data,
            "https://httpbin.org/status/429",
            use_failover=True,
            failover_timestamping_service_urls=[
                "https://httpbin.org/status/429",
                "http://freetsa.org/tsr",
            ],
            max_retries=2,
            initial_backoff_seconds=0.1,
        )
        self.assertIsInstance(tsr, bytes)
        self.assertIsInstance(timestamp_date, datetime)
        self.assertAlmostEqual(
            timestamp_date, datetime.now(pytz.UTC), delta=timedelta(seconds=10)
        )

    def test_timestamp_raises_connection_error(self):
        with self.assertRaises(ConnectionError):
            timestamp(
                self.file_data,
                "https://httpbin.org/status/500",
                use_failover=True,
                failover_timestamping_service_urls=["https://httpbin.org/status/429"],
                max_retries=2,
                initial_backoff_seconds=0.01,
            )

    def test_get_timestamp_date_from_tsr_file(self):
        tsr, timestamp_date = timestamp(self.file_data, "http://freetsa.org/tsr")
        verif_date = get_timestamp_date_from_tsr_file(tsr)
        self.assertEqual(timestamp_date, verif_date)
        self.assertAlmostEqual(
            verif_date, datetime.now(pytz.UTC), delta=timedelta(seconds=10)
        )



# -*- coding: utf-8 -*-

from collective.timestamp.behaviors.timestamp import ITimestampableDocument
from collective.timestamp.testing import COLLECTIVE_TIMESTAMP_FUNCTIONAL_TESTING
from plone import api
from plone.app.dexterity.behaviors.metadata import IBasic
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.behavior.interfaces import IBehavior
from plone.namedfile.file import NamedBlobFile
from plone.testing.z2 import Browser
from Products.statusmessages.interfaces import IStatusMessage
from rfc3161ng import TimestampingError
from unittest.mock import patch
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.interface import Interface
from zope.lifecycleevent import Attributes
from zope.lifecycleevent import modified

import transaction
import unittest


class TestBehavior(unittest.TestCase):

    layer = COLLECTIVE_TIMESTAMP_FUNCTIONAL_TESTING

    def setUp(self):
        """Custom shared utility setup for tests."""
        self.request = self.layer["request"]
        self.portal = self.layer["portal"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.document = api.content.create(
            container=self.portal,
            type="Document",
            id="my-document",
        )
        self.file = api.content.create(
            container=self.portal,
            type="File",
            id="my-file",
        )

    def test_behavior_interface(self):
        behavior = getUtility(IBehavior, "collective.timestamp")
        self.assertEqual(behavior.marker, ITimestampableDocument)
        self.assertTrue(ITimestampableDocument.providedBy(self.file))

    def test_action(self):
        view = getMultiAdapter((self.document, self.request), name="timestamp_utils")
        self.assertFalse(view.available())

        view = getMultiAdapter((self.file, self.request), name="timestamp_utils")
        self.assertFalse(view.available())
        self.file.file = NamedBlobFile(data=b"file data", filename="file.txt")
        self.assertTrue(view.available())

        view.timestamp()
        messages = IStatusMessage(self.request)
        show = messages.show()
        self.assertEqual(len(show), 1)
        self.assertIn(
            "Timestamp file has been successfully generated and saved", show[0].message
        )

        self.file.timestamp = None
        with patch(
            "collective.timestamp.adapters.TimeStamper.timestamp",
            side_effect=TimestampingError,
        ):
            view.timestamp()
            messages = IStatusMessage(self.request)
            show = messages.show()
            self.assertEqual(len(show), 2)
            self.assertIn("Timestamp has failed", show[1].message)

    def test_edition_warning(self):
        transaction.commit()
        browser = Browser(self.layer["app"])
        browser.addHeader(
            "Authorization",
            "Basic %s:%s"
            % (
                TEST_USER_NAME,
                TEST_USER_PASSWORD,
            ),
        )
        browser.open("{}/edit".format(self.file.absolute_url()))
        html = browser.contents
        self.assertNotIn(
            "You are editing a timestamped content.",
            html,
        )
        self.file.file = NamedBlobFile(data=b"file data", filename="file.txt")
        view = getMultiAdapter((self.file, self.request), name="timestamp_utils")
        view.timestamp()
        transaction.commit()
        browser.open("{}/edit".format(self.file.absolute_url()))
        html = browser.contents
        self.assertIn(
            "You are editing a timestamped content. Your modifications can invalidate the timestamp.",
            html,
        )

    def test_subscribers(self):
        self.file.file = NamedBlobFile(data=b"file data", filename="file.txt")
        modified(self.file, Attributes(IBasic, "IBasic.title"))
        messages = IStatusMessage(self.request)
        show = messages.show()
        self.assertEqual(len(show), 0)
        view = getMultiAdapter((self.file, self.request), name="timestamp_utils")
        self.assertTrue(view.available())
        view.timestamp()
        self.assertFalse(view.available())
        modified(self.file, Attributes(Interface, "file"))
        self.assertTrue(view.available())
        self.assertIsNone(self.file.timestamp)
        messages = IStatusMessage(self.request)
        show = messages.show()
        self.assertEqual(len(show), 2)
        self.assertIn(
            "Timestamp information has been removed since the data has changed",
            show[1].message,
        )



# -*- coding: utf-8 -*-

from collective.timestamp import _
from collective.timestamp.interfaces import ITimestampingSettings
from plone.app.registry.browser.controlpanel import ControlPanelFormWrapper
from plone.app.registry.browser.controlpanel import RegistryEditForm
from plone.z3cform import layout


class TimestampingControlPanelForm(RegistryEditForm):
    label = _("Timestamping settings")
    schema = ITimestampingSettings


TimestampingControlPanelView = layout.wrap_form(
    TimestampingControlPanelForm, ControlPanelFormWrapper
)



import time

import pytz
from collective.timestamp import logger
from datetime import datetime
from pyasn1.codec.der import decoder
from pyasn1.codec.der import encoder
from rfc3161ng import RemoteTimestamper
from rfc3161ng import TimeStampToken
from rfc3161ng import TimeStampResp
from rfc3161ng import get_timestamp


def localize_utc_date(date: datetime):
    tzinfo = pytz.timezone("UTC")
    return tzinfo.localize(date)


def get_timestamp_date(timestamp_token: TimeStampToken, localize=True):
    """
    Extract the timestamp date from a time stamp token.
    :param timestamp_token: The time stamp token to extract the date from.
    :param localize: Whether to localize the date to UTC (default is True).
    :return: The timestamp date, localized if specified.
    """
    timestamp_date = get_timestamp(timestamp_token)
    if localize:
        return localize_utc_date(timestamp_date)
    return timestamp_date


def get_timestamp_date_from_tsr_file(tsr_data: bytes, localize: bool = True):
    """Extracts the timestamp date from a tsr file"""
    tsr, _ = decoder.decode(tsr_data, asn1Spec=TimeStampResp())
    timestamp_token = tsr.time_stamp_token
    return get_timestamp_date(timestamp_token, localize=localize)


def timestamp(
    file_content: bytes,
    service_url: str,
    hashing_algorithm: str = "sha256",
    use_failover: bool = False,
    failover_timestamping_service_urls: list = None,
    max_retries: int = 0,
    initial_backoff_seconds: float = 0.5,
    localize_date: bool = True,
):
    """
    Generate a timestamp for the given file_content using a remote timestamping service.
    It attempts to obtain a timestamp from the (primary) service_url, and if it fails,
    it will switch to one of the failover_timestamping_service_urls if configured.
    It can also use exponential backoff for retries.
    :param file_content: The content of the file to be timestamped.
    :param service_url: The primary timestamping service URL.
    :param hashing_algorithm: The hashing algorithm to use (default is "sha256").
    :param use_failover: Whether to use failover timestamping services if the primary service fails.
    :param failover_timestamping_service_urls: A list of failover timestamping service URLs.
    :param max_retries: The maximum number of retries for timestamping attempts.
    :param initial_backoff_seconds: The initial backoff time in seconds for retrying failed attempts.
    :raises ConnectionError: If all attempts to obtain a timestamp fail.
    :return: A tuple containing the encoded timestamp response and the localized timestamp date.
    """
    success = False
    tsr = None

    while not success:
        retry_count = 0
        backoff_seconds = initial_backoff_seconds

        # Try main service URL with exponential backoff
        while retry_count <= max_retries:
            try:
                timestamper = RemoteTimestamper(
                    service_url,
                    certificate=b"",
                    hashname=hashing_algorithm,
                )
                tsr = timestamper(
                    data=file_content,
                    include_tsa_certificate=True,
                    return_tsr=True,
                )
                success = True
                break
            except Exception as e:
                retry_count += 1
                logger.error(f"Timestamping attempt {retry_count} failed: {e}")
                if retry_count <= max_retries:
                    logger.info(f"Retrying in {backoff_seconds} seconds...")
                    time.sleep(backoff_seconds)
                    backoff_seconds *= 2
                else:
                    logger.error(
                        f"Reached max retries ({max_retries}) for URL: {service_url}"
                    )

        if success or not use_failover:
            # If we succeeded or not using failover, we can exit the loop
            break

        if use_failover:
            if not failover_timestamping_service_urls:
                logger.error("No failover URLs available, cannot proceed.")
                break
            service_url = failover_timestamping_service_urls.pop(0)
            logger.info(f"Switching to failover URL: {service_url}")
    if not success:
        raise ConnectionError("Failed to obtain a timestamp.")

    timestamp_token = tsr.time_stamp_token
    timestamp_date = get_timestamp_date(timestamp_token, localize=localize_date)
    return encoder.encode(tsr), timestamp_date
